open Lf
open Printf

type ident = string

type concrete =
  | CKind
  | CType
  | CVar of ident
  | CProd of ident * concrete * concrete
  | CSig of ident * concrete * concrete
  | CSum of concrete * concrete
  | CEqu of concrete * concrete
  | CLam of ident * concrete * concrete
  | CApp of concrete * concrete
  | CPair of concrete * concrete * concrete
  | CPrj1 of concrete
  | CPrj2 of concrete
  | CInl of concrete * concrete
  | CInr of concrete * concrete
  | CCase of concrete * concrete * concrete
  | CRefl of concrete
  | CSubst of concrete * concrete * concrete * concrete * concrete

type program =
  | Const of (ident * concrete) * program
  | Def of (ident * concrete * concrete) * program
  | End

let rec count elt c =
  function
  | [] ->
      None
  | x :: l ->
      if x = elt then
	Some c
      else
	count elt (c+1) l

let ast_to_lf context env ast =
  let gamma = context.gamma in
  let delta = context.delta in
  let rec ast_to_lf env =
  function
  | CKind -> Kind
  | CType -> Type
  | CVar x ->
      if List.mem x (domain gamma) then
	Var x
      else if List.mem x (domain delta) then
	Var x
      else
	(match count x 0 env with
	 | None ->
	     failwith (sprintf "Variable %s not bound in %s." x (List.fold_left (fun acc x -> acc^" "^x) "" env))
	 | Some c ->
	     Rel c)
  | CProd(id,typ,body) ->
      Prod(ast_to_lf env typ,
	   ast_to_lf (id :: env) body)
  | CSig(id,typ,body) ->
      Sig(ast_to_lf env typ,
          ast_to_lf (id :: env) body)
  | CSum(l,r) ->
      Sum(ast_to_lf env l, ast_to_lf env r)
  | CEqu(l,r) ->
      Equ(ast_to_lf env l, ast_to_lf env r)
  | CLam(id,typ,body) ->
      Lam(ast_to_lf env typ,
	  ast_to_lf (id :: env) body)
  | CApp(f, a) ->
      App(ast_to_lf env f,
	  ast_to_lf env a)
  | CPair(t,l,r) ->
      Pair(ast_to_lf env t,
	   ast_to_lf env l,
	   ast_to_lf env r)
  | CPrj1 e -> 
      Prj1 (ast_to_lf env e)
  | CPrj2 e -> 
      Prj2 (ast_to_lf env e)
  | CInl(t, e) ->
      Inl(ast_to_lf env t, ast_to_lf env e)
  | CInr(t, e) ->
      Inr(ast_to_lf env t,ast_to_lf env e)
  | CCase(e,f,g) ->
      Case(ast_to_lf env e,
           ast_to_lf env f,
	   ast_to_lf env g)
  | CRefl e ->
      Refl(ast_to_lf env e)
  | CSubst(p,x,y,eq_xy,px) ->
      Subst(ast_to_lf env p,
            ast_to_lf env x,
            ast_to_lf env y,
            ast_to_lf env eq_xy,
            ast_to_lf env px)
  in
    ast_to_lf env ast
