open Printf
  
(* Term representation and type checking. *)

type const = string
type var   = string

(* Using the locally nameless representation. *)

type t =
  | Kind
  | Type
  | Var  of var
  | Rel  of int
  | Prod of t * t
  | Sig  of t * t
  | Sum  of t * t
  | Equ  of t * t
  | Lam  of t * t
  | App  of t * t
  | Pair of t * t * t
  | Prj1 of t
  | Prj2 of t
  | Inl  of t * t
  | Inr  of t * t
  | Case of t * t * t
  | Refl of t
  | Subst of t * t * t * t * t

(* context handling *)
type context = {
  gamma : (var * t) list;
  delta : (var * (t * t * context)) list
}

let empty_context =
  { gamma = [];
    delta = []; }

let context_add_var context var typ =
  { context with
    gamma = (var, typ) :: context.gamma }

let context_add_def context name proof typ =
  { context with
    delta = (name, (proof, typ, context)) :: context.delta }


type fresh_seed = int

let domain   gamma = List.map fst gamma
let codomain gamma = List.map snd gamma

let gensym : fresh_seed -> fresh_seed * var =
  fun seed ->
    (seed+1, "v"^(string_of_int seed))
      
let rec to_string fresh env expr =
  match expr with
  | Kind -> fresh, "Kind"
  | Type -> fresh, "Type"
  | Var v -> fresh, v
  | Rel i -> fresh, (List.nth env i)
  | Prod(tx,e) ->
      let fresh, txs = to_string fresh env tx in
      let fresh, x   = gensym fresh in
      let fresh, es  = to_string fresh (x::env) e in
      fresh, (sprintf "(forall (%s : %s), %s)" x txs es)
  | Sig(tx,e) ->
      let fresh, txs = to_string fresh env tx in
      let fresh, x   = gensym fresh in
      let fresh, es  = to_string fresh (x::env) e in
      fresh, (sprintf "(exists (%s : %s), %s)" x txs es)
  | Sum(l,r) ->
      let fresh, ls = to_string fresh env l in
      let fresh, rs = to_string fresh env r in
      fresh, (sprintf "(%s + %s)" ls rs)
  | Equ(x,y) ->
      let fresh, xs = to_string fresh env x in
      let fresh, ys = to_string fresh env y in
      fresh, (sprintf "(%s = %s)" xs ys)
  | Lam(tx,e) ->
      let fresh, txs = to_string fresh env tx in
      let fresh, x   = gensym fresh in
      let fresh, es  = to_string fresh (x::env) e in
      fresh, (sprintf "(fun (%s : %s) => %s)" x txs es)
  | App(a,b) ->
      let fresh, a_s = to_string fresh env a in
      let fresh, bs  = to_string fresh env b in
      fresh, (sprintf "(%s %s)" a_s bs)
  | Pair(tx,left,right) ->
      let fresh, ts = to_string fresh env tx in
      let fresh, ls = to_string fresh env left in
      let fresh, rs = to_string fresh env right in
      fresh, (sprintf "(%s,%s : %s)" ls rs ts)
  | Prj1 e ->
      let fresh, es = to_string fresh env e in
      fresh, (sprintf "(pi1 %s)" es)
  | Prj2 e ->
      let fresh, es = to_string fresh env e in
      fresh, (sprintf "(pi2 %s)" es)
  | Inl(t,e) ->
      let fresh, ts = to_string fresh env t in
      let fresh, es = to_string fresh env e in      
      fresh, (sprintf "(inj1 %s in %s)" es ts)
  | Inr(t,e) ->
      let fresh, ts = to_string fresh env t in
      let fresh, es = to_string fresh env e in      
      fresh, (sprintf "(inj2 %s in %s)" es ts)
  | Case(e, f, g) ->
      let fresh, es = to_string fresh env e in
      let fresh, fs = to_string fresh env f in
      let fresh, gs = to_string fresh env g in
      fresh, (sprintf "(case %s %s %s)" es fs gs)
  | Refl e ->
      let fresh, es = to_string fresh env e in
      fresh, (sprintf "(refl %s)" es)
  | Subst(p,x,y,eq_xy,px) ->
      let fresh, ps  = to_string fresh env p in
      let fresh, xs  = to_string fresh env x in
      let fresh, ys  = to_string fresh env y in
      let fresh, eqs = to_string fresh env eq_xy in
      let fresh, pxs = to_string fresh env px in
      fresh, (sprintf "(subst %s %s %s %s %s)" ps xs ys eqs pxs)

let rec fv expr acc =
  match expr with
  | Kind | Type -> acc
  | Var v -> v :: acc
  | Rel _ -> acc
  | Prod(tx,e)
  | Sig(tx,e)
  | Lam(tx,e) ->
      fv e (fv tx acc)
  | Sum(l,r) ->
      fv l (fv r acc)
  | Equ(x,y) ->
      fv x (fv y acc)
  | App(f,a) ->
      fv f (fv a acc)
  | Pair(tx, left, right) ->
      fv right (fv left (fv tx acc))
  | Prj1 e
  | Prj2 e ->
      fv e acc
  | Inl(t,e)
  | Inr(t,e) ->
      fv e (fv t acc)
  | Case(e,f,g) ->
      fv e (fv f (fv g acc))
  | Refl e ->
      fv e acc
  | Subst(p,x,y,eq_xy,px) ->
      fv p (fv x (fv y (fv eq_xy (fv px acc))))

(* Context printing. *)
let print_context fresh delta gamma term =
  let print_hyp fresh (v,t) =
    let fresh, s = to_string fresh [] t in
      fresh, (sprintf "%s : %s" v s)
  in
  let free_variables = fv term [] in
  let gamma' = List.filter (fun (v,_) -> List.mem v free_variables) gamma in
(* let delta' = List.filter (fun (v,_) -> (List.mem v free_variables) && (not (List.mem v (domain gamma')))) delta in *)
    List.fold_left (fun (fresh,str) elt ->
      let fresh, s = print_hyp fresh elt in
	fresh, (str^s^"\n")
    ) (fresh, "") gamma'

let print_problem fresh delta gamma term =
  let fresh,c = print_context fresh delta gamma term in
  let fresh,s = to_string fresh [] term in
    sprintf "%s\n----------------------------------------\n%s" c s

(* subst fresh var expr1 expr2 = fresh, expr2[var\expr1] *)
let subst_free var expr1 expr2 =
  let rec subst expr1 expr2 =
    match expr2 with
    | Kind | Type -> expr2
    | Var v ->
	if v = var then
	  expr1
	else
	  expr2
    | Rel _ -> expr2
    | Prod(tx,e) ->
	let tx' = subst expr1 tx in
	let e'  = subst expr1 e in
	Prod(tx', e')
    | Sig(tx,e) ->
	let tx' = subst expr1 tx in
	let e'  = subst expr1 e in
	Sig(tx', e')
    | Sum(l,r) ->
	let l'  = subst expr1 l in
	let r'  = subst expr1 r in
	Sum(l',r')
    | Equ(x,y) ->
	let x'  = subst expr1 x in
	let y'  = subst expr1 y in
	Equ(x',y')
    | Lam(tx,e) ->
	let tx' = subst expr1 tx in
	let e'  = subst expr1 e in
	Lam(tx', e')
    | App(f,a) ->
	let f'  = subst expr1 f in
	let a'  = subst expr1 a in
	App(f',a')
    | Pair(tx,left,right) ->
	let tx' = subst expr1 tx in
	let l'  = subst expr1 left in
	let r'  = subst expr1 right in
	Pair(tx',l',r')
    | Prj1 e ->
	let e' = subst expr1 e in
	Prj1 e'
    | Prj2 e ->
	let e' = subst expr1 e in
	Prj2 e'
    | Inl(t,e) ->
	let t' = subst expr1 t in
	let e' = subst expr1 e in
	Inl(t',e')
    | Inr(t,e) ->
	let t' = subst expr1 t in
	let e' = subst expr1 e in
	Inr(t',e')
    | Case(e,f,g) ->
	let e' = subst expr1 e in
	let f' = subst expr1 f in
	let g' = subst expr1 g in
	Case(e',f',g')
    | Refl e ->
	Refl(subst expr1 e)
    | Subst(p,x,y,eq_xy,px) ->
	let p'  = subst expr1 p in
	let x'  = subst expr1 x in
	let y'  = subst expr1 y in
	let eq' = subst expr1 eq_xy in
	let px' = subst expr1 px in
	  Subst(p',x',y',eq',px')
  in
  subst expr1 expr2


let rec rebind var deBruijn expr2 =
  let rec subst deBruijn expr2 =
  match expr2 with
    | Kind | Type -> expr2
    | Var v ->
	if v = var then
	  Rel deBruijn
	else
	  expr2
    | Rel _ -> expr2
    | Prod(tx,e) ->
	let tx' = subst deBruijn tx in
	let e'  = subst (deBruijn + 1) e in
	Prod(tx', e')
    | Sig(tx,e) ->
	let tx' = subst deBruijn tx in
	let e'  = subst (deBruijn + 1) e in
	Sig(tx', e')
    | Sum(l,r) ->
	let l'  = subst deBruijn l in
	let r'  = subst deBruijn r in
	Sum(l',r')
    | Equ(x,y) ->
	let x'  = subst deBruijn x in
	let y'  = subst deBruijn y in
	Equ(x',y')
    | Lam(tx,e) ->
	let tx' = subst deBruijn tx in
	let e'  = subst (deBruijn + 1) e in
	Lam(tx', e')
    | App(f,a) ->
	let f'  = subst deBruijn f in
	let a'  = subst deBruijn a in
	App(f',a')
    | Pair(tx,left,right) ->
	let tx' = subst deBruijn tx in
	let l'  = subst deBruijn left in
	let r'  = subst deBruijn right in
	Pair(tx',l',r')
    | Prj1 e ->
	let e' = subst deBruijn e in
	Prj1 e'
    | Prj2 e ->
	let e' = subst deBruijn e in
	Prj2 e'
    | Inl(t,e) ->
	let t' = subst deBruijn t in
	let e' = subst deBruijn e in
	Inl(t',e')
    | Inr(t,e) ->
	let t' = subst deBruijn t in
	let e' = subst deBruijn e in
	Inr(t',e')
    | Case(e,f,g) ->
	let e' = subst deBruijn e in
	let f' = subst deBruijn f in
	let g' = subst deBruijn g in
	Case(e',f',g')
    | Refl e ->
	Refl(subst deBruijn e)
    | Subst(p,x,y,eq_xy,px) ->
	let p'  = subst deBruijn p in
	let x'  = subst deBruijn x in
	let y'  = subst deBruijn y in
	let eq' = subst deBruijn eq_xy in
	let px' = subst deBruijn px in
	  Subst(p',x',y',eq',px')
	
  in
  subst deBruijn expr2
	

(* beta-reduction. *)
let subst_bound expr1 expr2 =
  let rec subst i expr1 expr2 =
    match expr2 with
    | Kind | Type -> expr2
    | Var _ -> expr2
    | Rel x ->
	if x = i then
	  expr1
	else
	  expr2
    | Prod(tx,e) ->
	let tx' = subst i expr1 tx in
	let e'  = subst (i+1) expr1 e in
	Prod(tx', e')
    | Sig(tx,e) ->
	let tx' = subst i expr1 tx in
	let e'  = subst (i+1) expr1 e in
	Sig(tx', e')
    | Sum(l,r) ->
	let l'  = subst i expr1 l in
	let r'  = subst i expr1 r in
	Sum(l',r')
    | Equ(x,y) ->
	let x'  = subst i expr1 x in
	let y'  = subst i expr1 y in
	  Equ(x',y')
    | Lam(tx,e) ->
	let tx' = subst i expr1 tx in
	let e'  = subst (i+1) expr1 e in
	Lam(tx', e')
    | App(f,a) ->
	let f'  = subst i expr1 f in
	let a'  = subst i expr1 a in
	App(f',a')
    | Pair(tx,left,right) ->
	let tx' = subst i expr1 tx in
	let l'  = subst i expr1 left in
	let r'  = subst i expr1 right in
	Pair(tx',l',r')
    | Prj1 e ->
	let e'  = subst i expr1 e in
	Prj1 e'
    | Prj2 e ->
	let e'  = subst i expr1 e in
	Prj2 e'
    | Inl(t,e) ->
	let t'  = subst i expr1 t in
	let e'  = subst i expr1 e in
	Inl(t',e')
    | Inr(t,e) ->
	let t'  = subst i expr1 t in
	let e'  = subst i expr1 e in
	Inr(t',e')
    | Case(e,f,g) ->
	let e'  = subst i expr1 e in
	let f'  = subst i expr1 f in
	let g'  = subst i expr1 g in
	Case(e',f',g')
    | Refl e ->
	Refl (subst i expr1 e)
    | Subst(p,x,y,eq_xy, px) ->
	let p'  = subst i expr1 p in
	let x'  = subst i expr1 x in
	let y'  = subst i expr1 y in
	let eq' = subst i expr1 eq_xy in
	let px' = subst i expr1 px in
	  Subst(p',x',y',eq',px')
  in
    subst 0 expr1 expr2

let rec normalize_cbn context expr =
  let delta = context.delta in
  let rec normalize expr =
    match expr with
      | Kind | Type -> expr
      | Var v -> 
	  if List.mem_assoc v delta then
	    let term, _, _ = List.assoc v delta in
	      term
	  else
	    expr
      | Rel _ -> failwith "Normalize (call by name) : relative index in term"
      | Prod(tx, body) -> Prod(normalize tx, body)
      | Sig(tx, body) -> Sig(normalize tx, body)
      | Sum(l, r) -> Sum(normalize l, normalize r)
      | Lam(tx, body) -> Lam(normalize tx, body)
      | Equ(x, y) -> Equ(normalize x, normalize y)
      | App(f, a) ->
	  (match normalize f with
	    | Lam(tx, body) ->
		normalize (subst_bound a body)
	    | x ->
		App(x, normalize a))
      | Pair(tx, left, right) ->
	  Pair(normalize tx,
          normalize left,
          normalize right)
      | Prj1 e ->
	  (match normalize e with
	    | Pair(_,x,_) -> x
	    | _ -> failwith "Normalize (call by name) : not a pair")
      | Prj2 e ->
	  (match normalize e with
	    | Pair(_,_,x) -> x
	    | _ -> failwith "Normalize (call by name) : not a pair")
      | Inl(t, e) ->
	  Inl(normalize t, normalize e)
      | Inr(t, e) ->
	  Inr(normalize t, normalize e)
      | Case(e,f,g) ->
	  (match normalize e with
	    | Inl(_, e) ->
		normalize (App(f,e))
	    | Inr(_, e) ->
		normalize (App(g,e))
	    | _ ->
		failwith "Normalize (call by name) : not a sum")
      | Refl e ->
	  Refl(normalize e)
      | Subst(p,x,y,eq_xy,px) ->
	  let x' = normalize x in
	  let y' = normalize y in
	  let e' = normalize eq_xy in
	    if x' = y' && e' = (Refl x') then
	      normalize px
	    else
	      Subst(normalize p, x', y', e', normalize px)
  in
    normalize expr


let fail_with_context fresh delta gamma term mess =
  let s = print_problem fresh delta gamma term in
    failwith ("\n"^s^"\n"^mess)
      
let typecheck fresh context term =
  let gamma = context.gamma in
  let delta = context.delta in
  let rec typecheck fresh gamma term =
  match term with
  | Kind ->
      fail_with_context fresh delta gamma term "While typing Kind : Kind has no type"
  | Type -> fresh, Kind
  | Var v ->
      if List.mem_assoc v gamma then
	fresh, (List.assoc v gamma)
      else if List.mem_assoc v delta then
	let _, typ, _ = List.assoc v delta in
	  fresh, typ
      else
	fail_with_context fresh delta gamma term "Unbound variable"
  | Rel _ ->
      fail_with_context fresh delta gamma term "De Bruijn variable encountered"
  | Prod(tx, body) ->
      let fresh, ttx = typecheck fresh gamma tx in
      if ttx = Type then
	let fresh, x = gensym fresh in
	let body     = subst_bound (Var x) body in
	let tx       = normalize_cbn context tx in
	let fresh, t = typecheck fresh ((x,tx) :: gamma) body in
	if t = Type || t = Kind then
	  fresh, t
	else
	  fail_with_context fresh delta gamma term "Codomain of product is not in Type or Kind"
      else
	let s = "While typing product : type of argument not in Type" in
	  fail_with_context fresh delta gamma term s
  | Sig(tx, body) ->
      let fresh, ttx = typecheck fresh gamma tx in
      if ttx = Type then
	let fresh, x = gensym fresh in
	let body     = subst_bound (Var x) body in
	let tx       = normalize_cbn context tx in
	let fresh, t = typecheck fresh ((x,tx) :: gamma) body in
	if t = Type then
	  fresh, t
	else
	  let s = "While typing sigma-type : right component not in Type" in
	    fail_with_context fresh delta gamma term s
      else
	let s = "While typing sigma-type : type of dependent variable not in Type" in
	  fail_with_context fresh delta gamma term s
  | Sum(l, r) ->
      let fresh, tl = typecheck fresh gamma l in
      let fresh, tr = typecheck fresh gamma r in
	if tl = Type && tr = Type then
	  fresh, Type
	else
	  let s = "While typing sum-type : left or right component not in Type" in
	    fail_with_context fresh delta gamma term s
  | Equ(x, y) ->
      let fresh, tx  = typecheck fresh gamma x in
      let fresh, ty  = typecheck fresh gamma y in
	if tx = ty then
	  let fresh, ttx = typecheck fresh gamma tx in
	  let fresh, tty = typecheck fresh gamma ty in
	    if ttx = Type && tty = Type then
	      fresh, Type
	    else
	      let s = "While typing equality-type : left or right component type not in Type" in
		fail_with_context fresh delta gamma term s
	else
	  let s = "While typing equality-type : components type mismatch" in
	    fail_with_context fresh delta gamma term s
  | Lam(tx, body) ->
      let fresh, ttx = typecheck fresh gamma tx in
      if ttx = Type then
	let fresh, x     = gensym fresh in
	let body         = subst_bound (Var x) body in
	let tx           = normalize_cbn context tx in
	let fresh, tbody = typecheck fresh ((x, tx) :: gamma) body in
	let tbody        = rebind x 0 tbody in
	fresh, (Prod(tx, tbody))
      else
	  let s = "While typing lambda-abstraction : type of argument is not of type Type" in
	    fail_with_context fresh delta gamma term s
  | App(f, a) ->
      let fresh, ft = typecheck fresh gamma f in
      let fresh, at = typecheck fresh gamma a in
      (match ft with
	| Prod(tx, body) ->
	    if tx = at then
	      let res = subst_bound a body in
		fresh, (normalize_cbn context res)
	    else
	      let s = "While typing application : type of argument does not match" in
		fail_with_context fresh delta gamma term s
	| _ ->
	    let s = "While typing application : not a product" in
	      fail_with_context fresh delta gamma term s)
  | Pair(tx, left, right) ->
      let fresh, lt  = typecheck fresh gamma left in
      let fresh, rt  = typecheck fresh gamma right in
      let tx         = normalize_cbn context tx in
      let fresh, ttx = typecheck fresh gamma tx in
      if ttx = Type then
	(match tx with
	 | Sig(lt', rt') ->
	     if (lt = lt') && (subst_bound (normalize_cbn context left) rt' = rt) then
	       fresh, tx
	     else
	       let s = "While typing dependent pair : type mismatch" in
		 fail_with_context fresh delta gamma term s
	 | _ ->
	     let s = "While typing dependent pair : not a sigma-type" in
	       fail_with_context fresh delta gamma term s)
      else
	let s = "While typing pair : pair type not in Type." in
	  fail_with_context fresh delta gamma term s	  
  | Prj1 e ->
      let fresh, et = typecheck fresh gamma e in
      (match et with
       | Sig(lt, _) ->
	   fresh, lt
       | _ ->
	   fail_with_context fresh delta gamma term "While typing pi1 : not a sigma-type")
  | Prj2 e ->
      let fresh, et = typecheck fresh gamma e in
      (match et with
       | Sig(lt, rt) ->
	   let res = subst_bound (Prj1 e) rt in
	   fresh, (normalize_cbn context res)
       | _ ->
	   fail_with_context fresh delta gamma term  "While typing pi2 : not a sigma-type")
  | Inl(t, e) ->
      let t = normalize_cbn context t in
	(match t with
	  | Sum(lt,rt) ->
	      let fresh, ltt = typecheck fresh gamma lt in
	      let fresh, rtt = typecheck fresh gamma rt in
		if ltt = Type && rtt = Type then
		  let fresh, et = typecheck fresh gamma e in
		    if et = lt then
		      fresh, t
		    else
		      let s = "While typing left injection : injected data type mismatch" in
			fail_with_context fresh delta gamma term s
		else
		  let s = "While typing left injection : components of sum-type not in Type." in
		    fail_with_context fresh delta gamma term s
	  | _ -> 
	      let s =  "While typing left injection : codomain not a sum-type." in
		fail_with_context fresh delta gamma term s)
  | Inr(t, e) ->
      let t = normalize_cbn context t in
	(match t with
	  | Sum(lt,rt) ->
	      let fresh, ltt = typecheck fresh gamma lt in
	      let fresh, rtt = typecheck fresh gamma rt in
		if ltt = Type && rtt = Type then
		  let fresh, et = typecheck fresh gamma e in
		    if et = rt then
		      fresh, t
		    else
		      let s = "While typing right injection : injected data type mismatch" in
			fail_with_context fresh delta gamma term s
		else
		  let s = "While typing right injection : components of sum-type not in Type." in
		    fail_with_context fresh delta gamma term s
	  | _ -> 
	      let s = "While typing right injection : codomain not a sum-type." in
		fail_with_context fresh delta gamma term s
	)
  | Case(e, f, g) ->
      let fresh, te = typecheck fresh gamma e in
      let fresh, tf = typecheck fresh gamma f in
      let fresh, tg = typecheck fresh gamma g in
	(match te with
	  | Sum(lt,rt) ->
	      (match tf, tg with
		| Prod(fdom,fcodom), Prod(gdom,gcodom) ->
		    if fdom = lt && gdom = rt then
		      if fcodom = gcodom then
			fresh, fcodom
		      else
			let s = "While typing case : codomain of right and left eliminator mismatch" in
			  fail_with_context fresh delta gamma term s
		    else
		      let s = "While typing case : domain of right or left eliminator mismatch" in
			fail_with_context fresh delta gamma term s
		| _ ->
		    let s = "While typing case : type of right or left eliminator not a product" in
		      fail_with_context fresh delta gamma term s)
	  | _ ->
	      let s = "While typing case : type of matched expression not a sum-type" in
		fail_with_context fresh delta gamma term s)
  | Refl e ->
      let fresh, te  = typecheck fresh gamma e in
      let fresh, tte = typecheck fresh gamma te in
	if tte = Type then
	  let e = normalize_cbn context e in
	  fresh, (Equ(e,e))
	else
	  let s = "While typing equality proof by reflectivity : left or right component type not in Type" in
	    fail_with_context fresh delta gamma term s
  | Subst(p,x,y,eq_xy,px) ->
      let fresh, tp  = typecheck fresh gamma p in
      let fresh, tx  = typecheck fresh gamma x in
      let fresh, ty  = typecheck fresh gamma y in
      let fresh, teq = typecheck fresh gamma eq_xy in
      let fresh, px  = typecheck fresh gamma px in
	(match tp with
	  | Prod(tx',Type) ->
	      if tx = ty then
		if tx = tx' then
		  match teq with
		    | Equ(tx'',ty'') ->
			let px' = normalize_cbn context (App(p,x)) in
			  if px' = (normalize_cbn context px) then
			    (* All seems ok. *)
			    fresh, (normalize_cbn context (App(p,y)))
			  else 
			    let s = "While typing substitution : invalid proposition proof" in
			      fail_with_context fresh delta gamma term s
		    | _ ->
			let s = "While typing substitution : not a valid equality proof" in
			  fail_with_context fresh delta gamma term s
		else
		  let s = "While typing substitution : domain of proposition does not match its arguments" in
		    fail_with_context fresh delta gamma term s
	      else
		let s = "While typing substitution : types of equality components mismatch" in
		  fail_with_context fresh delta gamma term s
	  | _ -> 
	      let s = "While typing substitution : proposition type not a product" in
		fail_with_context fresh delta gamma term s)
  in
    typecheck fresh gamma term
